@@ -190,15 +190,15 @@ class NormCipherFrame:
         if isinstance(other, NormCipherFrame):
             return NormCipherFrame(
                 self.apiContext,
-                self.cipherListOfListOfList+other.cipherListOfListOfList,
-                indexData = self.indicesListOfListOfList+other.indicesListOfListOfList,
-                dataTypes = self.dataTypes + other.dataTypes
+                deepcopy(self.cipherListOfListOfList)+deepcopy(other.cipherListOfListOfList),
+                indexData = deepcopy(self.indicesListOfListOfList)+deepcopy(other.indicesListOfListOfList),
+                dataTypes = deepcopy(self.dataTypes) + deepcopy(other.dataTypes)
             )
         elif isinstance(other,NormCipherList):
             newCipherList = deepcopy(self.cipherListOfListOfList)
-            newCipherList.append(other.cipherListOfList)
+            newCipherList.append(deepcopy(other.cipherListOfList))
             newIndicesList = deepcopy(self.indicesListOfListOfList)
-            newIndicesList.append(other.indicesListOfList)
+            newIndicesList.append(deepcopy(other.indicesListOfList))
             newDataTypes = deepcopy(self.dataTypes)
             newDataTypes.append('string')
             return NormCipherFrame(
@@ -229,15 +229,15 @@ class NormCipherFrame:
         return NormCipherFrame(
             self.apiContext,
             [
-                self.cipherListOfListOfList[i]+other.cipherListOfListOfList[i] 
+                deepcopy(self.cipherListOfListOfList[i])+deepcopy(other.cipherListOfListOfList[i])
                 for i in range(len(self.cipherListOfListOfList))
             ],
             indexData = [
-                self.indicesListOfListOfList[i]+other.indicesListOfListOfList[i]
+                deepcopy(self.indicesListOfListOfList[i])+deepcopy(other.indicesListOfListOfList[i])
                 for i in range(len(self.indicesListOfListOfList))
             ],
             dataTypes = self.dataTypes
-        )        
+        )          
     
     def vert_slice(self,rowList):
         return NormCipherFrame(
@@ -367,6 +367,56 @@ class NormCipherFrame:
             cipherFrame['id{}'.format(column)] = ids_to_append
     
         return cipherFrame.to_sql(table_name,connection)
+
+    def to_file(self,filename):
+        cipherLists = self.lol_of_ciphertext()
+        indexData = self.indicesListOfListOfList
+        if self.is_continuous():
+            indexData = (self.indicesListOfListOfList[0][0][0],self.indicesListOfListOfList[-1][-1][-1])
+        dataTypes = self.dataTypes
+        
+        
+        jsonToSave = json.dumps({'ciphers':cipherLists,'indices':indexData,
+                                 'dataTypes':dataTypes, 'colLengths': self.listOfColMaxChars})
+        
+        try:
+            with open(filename, "w") as outfile:
+                outfile.write(jsonToSave)
+        except:
+            raise Exception("Error writing to the file given")
+        
+        
+    def is_continuous(self):
+        lower = self.indicesListOfListOfList[0][0][0]
+        upper = self.indicesListOfListOfList[-1][-1][-1]
+        totalLength = sum([l*self.rows for l in self.listOfColMaxChars]) - 1 
+        
+        return lower + totalLength == upper
+    
+    def align_indices(self):
+        flatIndices = flatten_list(flatten_list(self.indicesListOfListOfList))
+        #print(flatIndices)
+        alignmentKey = align_index_key(self.apiContext,flatIndices)
+        minId = int(alignmentKey[0]['id'])
+        
+        newIndicesLOL = []
+        totalIndex = 0
+        for i in range(self.cols):
+            newIndicesL = []
+            for j in range(self.rows):
+                newIndicesL.append(list(range(minId,minId+self.listOfColMaxChars[i])))
+                for k in range(self.listOfColMaxChars[i]):
+                    self.cipherListOfListOfList[i][j][k] += int(alignmentKey[totalIndex]['atom_key'])
+                    if self.dataTypes[i] == 'float':
+                        self.cipherListOfListOfList[i][j][k] += int(alignmentKey[totalIndex]['atom_key_inv'])/32767.0
+                    totalIndex += 1
+                minId += self.listOfColMaxChars[i]
+            newIndicesLOL.append(newIndicesL)
+                
+                
+        self.indicesListOfListOfList = newIndicesLOL
+        
+        return self
     
  
     
@@ -456,4 +506,27 @@ def import_from_db(cryptoContext,table_name,connection):
         for j in range(indices.shape[1]):
             indices_list[i][j] = [k for k in range(indices[i][j],indices[i][j]+col_length)]
         meta_df = meta_df.drop('id{}'.format(i),axis=1)
-    return NormCipherFrame(cryptoContext,meta_df,indexData=indices_list)
\ No newline at end of file
+    return NormCipherFrame(cryptoContext,meta_df,indexData=indices_list)
+
+
+def import_from_file(cryptoContext,filename):
+    with open(filename, 'r') as f:
+        jsonDict = json.load(f)
+        
+    cipherListOfListOfList = jsonDict['ciphers']
+    indicesListOfListOfList = jsonDict['indices']
+    if isinstance(indicesListOfListOfList[0], int):
+        lower = min(indicesListOfListOfList)
+        upper = max(indicesListOfListOfList)
+        indicesListOfListOfList = []
+        colLengths = jsonDict['colLengths']
+        numRows = len(cipherListOfListOfList[0])
+        prevIndex = lower
+        for colIndex in range(len(cipherListOfListOfList)):
+            indicesListOfList = [list(range(j,j+colLengths[colIndex]))
+                                 for j in range(prevIndex,prevIndex+colLengths[colIndex]*numRows,colLengths[colIndex])]
+            indicesListOfListOfList.append(indicesListOfList)
+            prevIndex += colLengths[colIndex]*numRows
+    cipherListOfListOfList = [[decode_ciphertext(word) for word in cipherList] for cipherList in cipherListOfListOfList]
+    return NormCipherFrame(cryptoContext,cipherListOfListOfList,
+                           indexData=indicesListOfListOfList,dataTypes=jsonDict['dataTypes'])
\ No newline at end of file
